/*
 *  LED_Synch_Mesh_Combined_Neighborhood
 *  Each ESP8266 based mesh node has the same number of WS2811 ("Neopixel") LEDs on pin D6
 *  The animations are synched between all the nodes
 *  
 *  On the shoulders of giants - uses PainlessMesh and FastLED
 *  
 *  2020 Carl F Sutter
 *  
 */

/* !!!!!! IMPORTANT: This is used to set the root node - the one that controls the other ones.
 *  This is used if you have a lot or they are spread out and not synchronizing - you set one to the root
 *  and all the others will just synch to that one
 *  https://gitlab.com/painlessMesh/painlessMesh/-/wikis/Root-nodes
 *  So, set this to "true" for one main node, and "false" for all the others!
*/
#define bISROOT false


// LED Setup - must come before the painlessMesh so CRGB can be used as a return type
#include <FastLED.h>
// How many leds in your strip?
#define NUM_LEDS 99

// For led chips like Neopixels, which have a data line, ground, and power, you just
// need to define DATA_PIN.  Note that the GPIO and Data pins are laballed separately on ESP8266 boards
#define DATA_PIN D6

// Define the array of leds
CRGB leds[NUM_LEDS];


// Global Vars and constants
#define MODE_SINGLE 1

int display_mode;                // this is the current animation type
uint16_t display_step;           // this is a count of the steps in the current animation
//bool amController = false;       // flag to designate that this CPU is the current controller

#define MODE6CLICKSPERCOLOR 10   // the number of 50us counts to hold one color
//#define MODE6NUMLOOPS 8          // the number of times to change the color
#define MODE6NUMLOOPS 4          // the number of times to change the color

unsigned long ul_PreviousMillis = 0UL;   // using millis rollver code from http://playground.arduino.cc/Code/TimingRollover
unsigned long ul_Interval;               // this is delay between animation steps ms

unsigned long ul_TimeoutPreviousMillis = 0UL;   // starting time
unsigned long ul_TimeoutInterval = 600000UL;     // 10 second delay to see if we have any commands



// Mesh Setup
#include "painlessMesh.h"

#define   MESH_PREFIX     "HolidayLEDs"
#define   MESH_PASSWORD   "foobar123"       // needs to be at leats 8 characters!
#define   MESH_PORT       5555

void sendMessage(int display_step); // Prototype so PlatformIO doesn't complain

painlessMesh  mesh;
//Task taskSendMessage( TASK_SECOND * 1 , TASK_FOREVER, &sendMessage );


// this gets called when the designated controller sends a command to start a new animation
// init any animation specific vars for the new mode, and reset the timer vars
void receivedCallback( uint32_t from, String &msg ) {
  // get the new display mode
  display_mode = msg.toInt();

  // init some vars that are the same for all animations
  display_step = 0;
  ul_PreviousMillis = 0UL; // make sure animation triggers on first step
  for(uint16_t i=0; i<NUM_LEDS; i++) leds[i] = CRGB(0, 0, 0);  // clear the LEDS

  // start the no command timer
  ul_TimeoutPreviousMillis = millis();
} // receivedCallback

/*
// called on a new connection - ignore since we will used changedConnection
void newConnectionCallback(uint32_t nodeId) {
    Serial.printf("--> startHere: New Connection, nodeId = %u\n", nodeId);
} // newConnectionCallback


// this gets called when a node is added or removed from the mesh, so set the controller to the node with the lowest chip id
void changedConnectionCallback() {
  SimpleList<uint32_t> nodes;
  uint32_t myNodeID = mesh.getNodeId();
  uint32_t lowestNodeID = myNodeID;

  Serial.printf("Changed connections %s\n",mesh.subConnectionJson().c_str());

  nodes = mesh.getNodeList();
  Serial.printf("Number of other nodes: %d\n", nodes.size());
  Serial.printf("Connection list:");

  for (SimpleList<uint32_t>::iterator node = nodes.begin(); node != nodes.end(); ++node) {
    Serial.printf(" %u", *node);
    if (*node < lowestNodeID) lowestNodeID = *node;
  }
  Serial.println();

  if (lowestNodeID == myNodeID) {
    Serial.printf("Node %u: I am the controller now", myNodeID);
    Serial.println();
    amController = true;
    // restart the current animation - to chatty - remove - better to wait for next animation
    //sendMessage(display_mode);
    //display_step = 0;
    //ul_PreviousMillis = 0UL; // make sure animation triggers on first step
  } else {
    Serial.printf("Node %u is the controller now", lowestNodeID);
    Serial.println();
    amController = false;
  }
} // changedConnectionCallback
*/

// time adjust - not using for this project
void nodeTimeAdjustedCallback(int32_t offset) {
    Serial.printf("Adjusted time %u. Offset = %d\n", mesh.getNodeTime(),offset);
} // changedConnectionCallback


// setup
void setup() {
  FastLED.addLeds<WS2812B, DATA_PIN, RGB>(leds, NUM_LEDS);
  FastLED.setBrightness(60);
  
  Serial.begin(115200);

//mesh.setDebugMsgTypes( ERROR | MESH_STATUS | CONNECTION | SYNC | COMMUNICATION | GENERAL | MSG_TYPES | REMOTE ); // all types on
  mesh.setDebugMsgTypes( ERROR | STARTUP );  // set before init() so that you can see startup messages

  mesh.init( MESH_PREFIX, MESH_PASSWORD, MESH_PORT );
  mesh.onReceive(&receivedCallback);
  //mesh.onNewConnection(&newConnectionCallback);
  //mesh.onChangedConnections(&changedConnectionCallback);
  //mesh.onNodeTimeAdjusted(&nodeTimeAdjustedCallback);
  // Bridge node, should (in most cases) be a root node. See [the wiki](https://gitlab.com/painlessMesh/painlessMesh/wikis/Possible-challenges-in-mesh-formation) for some background
  if (bISROOT) mesh.setRoot(true);
  // This node and all other nodes should ideally know the mesh contains a root, so call this on all nodes
  mesh.setContainsRoot(true);

  Serial.print("My Node ID: ");
  Serial.println(mesh.getNodeId());

  //mesh.scheduler.addTask( taskSendMessage );
  //taskSendMessage.enable() ;

  display_mode = MODE_SINGLE;
  sendMessage(display_mode);
  display_step = 0;

  // make this one the controller if there are no others on the mesh
  //if (mesh.getNodeList().size() == 0) amController = true;

  // start the no command timer
  ul_TimeoutPreviousMillis = millis();
} // setup


// loop
void loop() {
  mesh.update();
  stepAnimation();  // update any animation if it's time

  // reset the CPU if we have had no commands in 10 seconds
  unsigned long ul_CurrentMillis = millis();  
  if (ul_CurrentMillis - ul_TimeoutPreviousMillis > ul_TimeoutInterval) {
    //ul_TimeoutPreviousMillis = millis();  // temporary just reset
    ESP.restart();  // normal reboot 
    //ESP.reset();  // hard reset
  } 
} // loop


// sort the given list of nodes
void sortNodeList(SimpleList<uint32_t> &nodes) {
  SimpleList<uint32_t> nodes_sorted;
  SimpleList<uint32_t>::iterator smallest_node;

  // sort the node list
  while (nodes.size() > 0) {
    // find the smallest one
    smallest_node = nodes.begin();
    for (SimpleList<uint32_t>::iterator node = nodes.begin(); node != nodes.end(); ++node) {
      if (*node < *smallest_node) smallest_node = node;
    }
    
    // add it to the sorted list and remove it from the old list
    nodes_sorted.push_back(*smallest_node);
    nodes.erase(smallest_node);
  } // while
  
  // copy the sorted list back into the now empty nodes list
  for (SimpleList<uint32_t>::iterator node = nodes_sorted.begin(); node != nodes_sorted.end(); ++node) nodes.push_back(*node);
} // sortNodeList


// check the timer and do one animation step if needed
void stepAnimation() {
  bool display_mode_changed = false;
  int even_increment = 0;

  // if this is the first step of the animation, set the display animation interval
  if (display_step == 0) {
    ul_Interval = 50UL;  // almost all of them use 50ms
    if (display_mode == 3) ul_Interval = 5UL;
  }
  
  // if the animation delay time has past, run another animation step
  unsigned long ul_CurrentMillis = millis();  
  if (ul_CurrentMillis - ul_PreviousMillis > ul_Interval) {
    ul_PreviousMillis = millis();
    //Serial.print(display_mode);
    switch (display_mode) {
      case MODE_SINGLE: // run the single color up the bar one LED at a time
        if (display_step == 0) for(uint16_t i=0; i<NUM_LEDS; i++) leds[i] = CRGB(0, 255, 0); // set all the strips to green
        leds[display_step % NUM_LEDS] = CRGB(255, 0, 0);
        FastLED.show();

        // check if we are done
        display_step++;
        if (display_step > NUM_LEDS) {
          display_mode_changed = true;
          //display_mode = 2;  // show all the animations
          //display_mode = 3;  // skip to the holiday only ones
          display_mode = 4;  // skip to the holiday only ones
        }        
      break;

      case 2: // rainbow colors
        leds[display_step % NUM_LEDS] = Wheel(display_step % 255);
        // we could also set all the pixels at once
        //for(uint16_t i=0; i< strip.numPixels(); i++) {
        //  strip.setPixelColor(i, Wheel((i+display_step) & 255));
        //  //strip.setPixelColor(i, Wheel(((i * 256 / strip.numPixels()) + display_step) & 255));
        //}
        FastLED.show();

        // check if we are done
        display_step++;
        if (display_step > 255) {
          display_mode_changed = true;
          display_mode = 3;
        }
      break;

      case 3: // go from dim to full on one color for all LEDs a few times
        for(uint16_t i=0; i<NUM_LEDS; i++) {
          leds[i] = CRGB(0, 0, display_step % 255);
        }
        FastLED.show();

        // check if we are done
        display_step++;
        if (display_step > (255 * 5)) {
          display_mode_changed = true;
          display_mode = 4;
        }
      break;

      case 4: // alternate red and green LEDs
        if ((NUM_LEDS % 2) == 0) even_increment = display_step / NUM_LEDS;
        //if (((display_step + (display_step / NUM_LEDS)) % 2) == 0) {
        //if ((display_step % 2) == 0) {
        if (((display_step + even_increment) % 2) == 0) {
          leds[display_step % NUM_LEDS] = CRGB(255, 0, 0);
        } else {
          leds[display_step % NUM_LEDS] = CRGB(0, 255, 0);
        }
        FastLED.show();

        // check if we are done
        display_step++;
        if (display_step > (NUM_LEDS*4)) {
          display_mode_changed = true;
          //display_mode = 5;
          display_mode = 6;  // skip the dot moving one
        }
      break;

      case 5: // animate a dot moving up all of them 
      {
        int curr_strip_num = display_step / NUM_LEDS;
        SimpleList<uint32_t> nodes = mesh.getNodeList();
        int node_index = 0;
        uint32_t myNodeID = mesh.getNodeId();

        // add this node to the node list and sort the list
        nodes.push_back(myNodeID);
        sortNodeList(nodes);
        
        // blacken all the leds
        for(uint16_t i=0; i<NUM_LEDS; i++) leds[i] = CRGB(0, 0, 0);

        // if this is the active strip, set the next LED on
        SimpleList<uint32_t>::iterator node = nodes.begin();
        while (node != nodes.end()) {
          if ((*node == myNodeID) && (node_index == curr_strip_num)) {
            leds[display_step % NUM_LEDS] = CRGB(255, 255, 255);
          }
          node_index++;
          node++;
        }
        FastLED.show();
      }

        // check if we are done
        display_step++;
        if (display_step >= ((mesh.getNodeList().size() + 1) * NUM_LEDS)) {
          display_mode_changed = true;
          display_mode = 6;
        }
      break;

      case 6: // green bars with a red bar moving around them a few times
      {
        SimpleList<uint32_t> nodes = mesh.getNodeList();
        int curr_strip_num = (display_step / MODE6CLICKSPERCOLOR) % (nodes.size() + 1);
        int node_index = 0;
        uint32_t myNodeID = mesh.getNodeId();

        // add this node to the node list and sort the list
        nodes.push_back(myNodeID);
        sortNodeList(nodes);

        // set all the leds to green
        for(uint16_t i=0; i<NUM_LEDS; i++) leds[i] = CRGB(0, 255, 0);

        // if this is the active strip, set the next LED on
        SimpleList<uint32_t>::iterator node = nodes.begin();
        while (node != nodes.end()) {
          if ((*node == myNodeID) && (node_index == curr_strip_num)) {
            for(uint16_t i=0; i<NUM_LEDS; i++) leds[i] = CRGB(255, 0, 0);
          }
          node_index++;
          node++;
        }
        FastLED.show();
      }

        // check if we are done
        display_step++;
        if (display_step >= ((mesh.getNodeList().size() + 1) * MODE6CLICKSPERCOLOR * MODE6NUMLOOPS)) {  // 10 steps per color (x50us), 8 color changes.  total run time is 8*(#strips/2)
          display_mode_changed = true;
          //display_mode = MODE_SINGLE;
          display_mode = 8;
        }
      break;
      
      case 7: // white random light up
        //if (display_step == 0) for(uint16_t i=0; i<NUM_LEDS; i++) leds[i] = CRGB(0, 255, 0); // set all the strips to green
        //leds[display_step % NUM_LEDS] = CRGB(255, 255, 255);
        leds[random(NUM_LEDS)] = CRGB(255, 255, 255);
        FastLED.show();

        // check if we are done
        display_step++;
        if (display_step > (NUM_LEDS * 2)) {
          display_mode_changed = true;
          //display_mode = 2;  // show all the animations
          display_mode = MODE_SINGLE;
        }        
      break;

      case 8: // white dot moving up
        // blacken all the leds
        for(uint16_t i=0; i<NUM_LEDS; i++) leds[i] = CRGB(0, 0, 0);

        // light up the current one - allow for multipe passes
        leds[display_step % NUM_LEDS] = CRGB(255, 255, 255);

        FastLED.show();

        // check if we are done - 2x through
        display_step++;
        if (display_step > (NUM_LEDS * 2)) {
          display_mode_changed = true;
          //display_mode = 2;  // show all the animations
          display_mode = 9;
        }        
      break;

      case 9: // xmas colors
        //if (display_step == 0) {
          // blacken all the leds
          //for(uint16_t i=0; i<NUM_LEDS; i++) leds[i] = CRGB(0, 0, 0);

          // light up the 5 colors
          for(uint16_t i=0; i<NUM_LEDS; i+=5) {
            leds[i]   = CRGB(255,   0,   0);
            leds[i+1] = CRGB(  0, 255,   0);
            leds[i+2] = CRGB(204,  50,   0);
            leds[i+3] = CRGB(  0,   0, 255);
            leds[i+4] = CRGB(255, 245,   0);
          }
          FastLED.show();
        //} // display_step == 0

        // check if we are done
        display_step++;
        if (display_step > (NUM_LEDS * 4)) {
          display_mode_changed = true;
          //display_mode = 2;  // show all the animations
          display_mode = 10;
        }        
      break;

      case 10: // alternate white and blue LEDs
        if ((NUM_LEDS % 2) == 0) even_increment = display_step / NUM_LEDS;
        //if (((display_step + (display_step / NUM_LEDS)) % 2) == 0) {
        //if ((display_step % 2) == 0) {
        if (((display_step + even_increment) % 2) == 0) {
          leds[display_step % NUM_LEDS] = CRGB(255, 255, 255);
        } else {
          leds[display_step % NUM_LEDS] = CRGB(0, 0, 255);
        }
        FastLED.show();

        // check if we are done
        display_step++;
        if (display_step > (NUM_LEDS*4)) {
          display_mode_changed = true;
          //display_mode = 5;
          display_mode = MODE_SINGLE;
        }
      break;

    } // display_mode switch

    // if we are done and the display mode changed, do some common tasks and set some common vars
    // only the controller does this, and it messages the rest
    if (display_mode_changed) {
      display_step = 0;
      ul_PreviousMillis = 0UL; // make sure animation triggers on first step
      for(uint16_t i=0; i<NUM_LEDS; i++) leds[i] = CRGB(0, 0, 0);  // clear the LEDS

      //if (amController) sendMessage(display_mode);
      if (bISROOT) sendMessage(display_mode);
    }

    // a debug led to show which is the current controller
    //if (amController) leds[0] = CRGB(64, 64, 64);

  } // timing conditional ul_Interval
} // stepAnimation


// send a broadcast message to all the nodes specifying the new animation mode for all of them
void sendMessage(int display_mode) {
  Serial.print("Sending message to change display_mode to ");
  Serial.println(display_mode);
  String msg;
  //msg += mesh.getNodeId();
  msg += String(display_mode);
  mesh.sendBroadcast( msg );
} // sendMessage


// Input a value 0 to 255 to get a color value.
// The colours are a transition r - g - b - back to r.
CRGB Wheel(byte WheelPos) {
  WheelPos = 255 - WheelPos;
  if(WheelPos < 85) {
    return CRGB(255 - WheelPos * 3, 0, WheelPos * 3);
  }
  if(WheelPos < 170) {
    WheelPos -= 85;
    return CRGB(0, WheelPos * 3, 255 - WheelPos * 3);
  }
  WheelPos -= 170;
  return CRGB(WheelPos * 3, 255 - WheelPos * 3, 0);
}
